import math

from tests.v1.test_base import TestBase


class TestJobVertexSubtaskClient(TestBase):

    def test_accumulators(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        jvc = simple_client.jobs.get_vertex(jid, vid).subtasks

        requests_mock.get(f'{jvc.prefix_url}/accumulators', json={
            'id': 'bc764cd8ddf7a0cff126f51c16239658',
            'parallelism': 1,
            'subtasks': [{'subtask': 0, 'attempt': 0, 'host': 'flink-taskmanager', 'user-accumulators': []}]})
        response = jvc.accumulators()

        assert isinstance(response, dict)
        assert response['id'] == 'bc764cd8ddf7a0cff126f51c16239658'
        assert response['parallelism'] == 1
        assert isinstance(response['subtasks'], list)

    def test_subtask_ids(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        jvc = simple_client.jobs.get_vertex(jid, vid).subtasks

        requests_mock.get(f'{jvc.prefix_url}/accumulators', json={
            'id': 'bc764cd8ddf7a0cff126f51c16239658',
            'parallelism': 1,
            'subtasks': [{'subtask': 0, 'attempt': 0, 'host': 'flink-taskmanager', 'user-accumulators': []}]})
        response = jvc.subtask_ids()

        assert isinstance(response, list)
        assert len(response) == 1
        assert response[0] == 0

    def test_metric_names(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        jvc = simple_client.jobs.get_vertex(jid, vid).subtasks

        requests_mock.get(f'{jvc.prefix_url}/metrics', json=[
            {'id': 'Shuffle.Netty.Input.numBytesInLocal'},
            {'id': 'Shuffle.Netty.Output.Buffers.outPoolUsage'},
            {'id': 'checkpointStartDelayNanos'},
        ])
        response = jvc.metric_names()

        assert isinstance(response, list)
        assert len(response) == 3
        assert response[0] == 'Shuffle.Netty.Input.numBytesInLocal'
        assert response[1] == 'Shuffle.Netty.Output.Buffers.outPoolUsage'
        assert response[2] == 'checkpointStartDelayNanos'

    def test_metrics(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        jvc = simple_client.jobs.get_vertex(jid, vid).subtasks

        requests_mock.get(f'{jvc.prefix_url}/accumulators', json={
            'id': 'bc764cd8ddf7a0cff126f51c16239658',
            'parallelism': 1,
            'subtasks': [{'subtask': 0, 'attempt': 0, 'host': 'flink-taskmanager', 'user-accumulators': []}]})

        requests_mock.get(f'{jvc.prefix_url}/metrics', json=[
            {'id': 'Shuffle.Netty.Input.numBytesInLocal',
             'min': 0.0, 'max': 0.0, 'avg': 0.0, 'sum': 0.0},
            {'id': 'Shuffle.Netty.Output.Buffers.outPoolUsage',
             'min': 8482.0, 'max': 8482.0, 'avg': 8482.0, 'sum': 8482.0},
        ])
        response = jvc.metrics()

        assert isinstance(response, dict)
        assert math.isclose(response['Shuffle.Netty.Input.numBytesInLocal']['min'], 0.0)
        assert math.isclose(response['Shuffle.Netty.Output.Buffers.outPoolUsage']['min'], 8482.0)

    def test_get(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        subtask_id = 0
        jvc = simple_client.jobs.get_vertex(jid, vid).subtasks

        requests_mock.get(f'{jvc.prefix_url}/{subtask_id}', json={
            'subtask': 0,
            'status': 'RUNNING',
            'attempt': 0,
            'host': 'flink-taskmanager',
            'start-time': 101010402, 'end-time': -1, 'duration': 2136900,
            'metrics': {'read-bytes': 0, 'read-bytes-complete': True, 'write-bytes': 7862928,
                        'write-bytes-complete': True, 'read-records': 0, 'read-records-complete': True,
                        'write-records': 564821, 'write-records-complete': True},
            'taskmanager-id': '172.18.0.3:40027-9e32e4',
            'start_time': 101010402})
        response = jvc.get(subtask_id)

        assert isinstance(response, dict)
        assert response['status'] == 'RUNNING'
        assert response['taskmanager-id'] == '172.18.0.3:40027-9e32e4'

    def test_get_attempt(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        subtask_id = 0
        jvc = simple_client.jobs.get_vertex(jid, vid).subtasks

        requests_mock.get(f'{jvc.prefix_url}/{subtask_id}', json={
            'subtask': 0,
            'status': 'RUNNING',
            'attempt': 0,
            'host': 'flink-taskmanager',
            'start-time': 101010402, 'end-time': -1, 'duration': 2136900,
            'metrics': {'read-bytes': 0, 'read-bytes-complete': True, 'write-bytes': 7862928,
                        'write-bytes-complete': True, 'read-records': 0, 'read-records-complete': True,
                        'write-records': 564821, 'write-records-complete': True},
            'taskmanager-id': '172.18.0.3:40027-9e32e4',
            'start_time': 101010402})

        requests_mock.get(f'{jvc.prefix_url}/{subtask_id}/attempts/{subtask_id}', json={
            'subtask': 0,
            'status': 'RUNNING',
            'attempt': 0,
            'host': 'flink-taskmanager',
            'start-time': 101010402, 'end-time': -1, 'duration': 2136900,
            'metrics': {'read-bytes': 0, 'read-bytes-complete': True, 'write-bytes': 7862928,
                        'write-bytes-complete': True, 'read-records': 0, 'read-records-complete': True,
                        'write-records': 564821, 'write-records-complete': True},
            'taskmanager-id': '172.18.0.3:40027-9e32e4',
            'start_time': 101010402})
        response = jvc.get_attempt(subtask_id)

        assert isinstance(response, dict)
        assert response['status'] == 'RUNNING'
        assert response['taskmanager-id'] == '172.18.0.3:40027-9e32e4'

    def test_get_attempt_accumulators(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        subtask_id = 0
        jvc = simple_client.jobs.get_vertex(jid, vid).subtasks

        requests_mock.get(f'{jvc.prefix_url}/{subtask_id}', json={
            'subtask': 0,
            'status': 'RUNNING',
            'attempt': 0,
            'host': 'flink-taskmanager',
            'start-time': 101010402, 'end-time': -1, 'duration': 2136900,
            'metrics': {'read-bytes': 0, 'read-bytes-complete': True, 'write-bytes': 7862928,
                        'write-bytes-complete': True, 'read-records': 0, 'read-records-complete': True,
                        'write-records': 564821, 'write-records-complete': True},
            'taskmanager-id': '172.18.0.3:40027-9e32e4',
            'start_time': 101010402})

        requests_mock.get(f'{jvc.prefix_url}/{subtask_id}/attempts/{subtask_id}/accumulators', json={
            'subtask': 0,
            'attempt': 0,
            'id': 'b0a52b89389fe646b47eb9632bd7f1f8',
            'user-accumulators': []})
        response = jvc.get_attempt_accumulators(subtask_id)

        assert isinstance(response, dict)
        assert response['subtask'] == 0
        assert response['attempt'] == 0
        assert response['id'] == 'b0a52b89389fe646b47eb9632bd7f1f8'



from flink_rest_client.common import RestException
from flink_rest_client.v1.client import FlinkRestClientV1

VERSIONS = {"v1": FlinkRestClientV1}


class FlinkRestClient:
    @staticmethod
    def get(host, port=None, version=None):
        """
        Constructs a new rest client instance.

        Parameters
        ----------
        host: str
            Hostname of Flink Jobmanager
        port: int
            Port number. Default value: 8081
        version: str
            Version of the REST API. Default value: v1
        version
        """
        port = 8081 if port is None else port
        version = "v1" if version is None else version

        if version not in VERSIONS.keys():
            raise RestException(f"Unknown REST API version: {version}")
        api_client_cls = VERSIONS[version]
        return api_client_cls(host=host, port=port)




from flink_rest_client.v1.client import DatasetTrigger
from tests.v1.test_base import TestBase


class TestClientV1(TestBase):

    def test_api_url(self, simple_client):
        assert simple_client.api_url == 'http://host:8081/v1'

    def test_jobmanager(self, simple_client):
        assert simple_client.jobmanager.prefix == 'http://host:8081/v1/jobmanager'

    def test_taskmanagers(self, simple_client):
        assert simple_client.taskmanagers.prefix == 'http://host:8081/v1/taskmanagers'

    def test_jars(self, simple_client):
        assert simple_client.jars.prefix == 'http://host:8081/v1/jars'

    def test_overview(self, simple_client, requests_mock):
        requests_mock.get(f'{simple_client.api_url}/overview', json={
            'taskmanagers': 1,
            'flink-version': '1.12.4'
        })
        response = simple_client.overview()

        assert isinstance(response, dict)
        assert response['taskmanagers'] == 1
        assert response['flink-version'] == '1.12.4'

    def test_config(self, simple_client, requests_mock):
        requests_mock.get(f'{simple_client.api_url}/config', json={
            'refresh-interval': 3000,
            'flink-version': '1.12.4'
        })
        response = simple_client.config()

        assert isinstance(response, dict)
        assert response['refresh-interval'] == 3000
        assert response['flink-version'] == '1.12.4'

    def test_delete_cluster(self, simple_client, requests_mock):
        requests_mock.delete(f'{simple_client.api_url}/cluster', json={})
        response = simple_client.delete_cluster()

        assert isinstance(response, dict)
        assert len(response) == 0

    def test_datasets(self, simple_client, requests_mock):
        requests_mock.get(f'{simple_client.api_url}/datasets', json={
            'dataSets': []
        })
        response = simple_client.datasets()

        assert isinstance(response, list)
        assert len(response) == 0

    def test_delete_dataset(self, simple_client, requests_mock):
        dataset_id = 'test_id'
        trigger_id = 5555
        requests_mock.delete(f'{simple_client.api_url}/datasets/{dataset_id}', json={
            'request-id': trigger_id
        }, status_code=202)
        response = simple_client.delete_dataset(dataset_id)

        assert isinstance(response, DatasetTrigger)
        assert response.trigger_id == trigger_id

        requests_mock.get(f'{simple_client.api_url}/datasets/delete/{trigger_id}', json={
            'status': 'finished'
        })
        status = response.status
        assert isinstance(status, dict)
        assert status['status'] == 'finished'






import requests


class RestException(Exception):
    """
    Exception to catch REST API related exceptions.
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def _execute_rest_request(
    url,
    http_method=None,
    accepted_status_code=None,
    files=None,
    params=None,
    data=None,
    json=None,
):
    if http_method is None:
        http_method = "GET"
    if params is None:
        params = {}
    if data is None:
        data = {}

    # If accepted_status_code is None then default value is set.
    if accepted_status_code is None:
        accepted_status_code = 200

    response = requests.request(
        method=http_method, url=url, files=files, params=params, data=data, json=json
    )
    if response.status_code == accepted_status_code:
        return response.json()
    else:
        if "errors" in response.json().keys():
            error_str = "\n".join(response.json()["errors"])
        else:
            error_str = ""
        raise RestException(
            f"REST response error ({response.status_code}): {error_str}"
        )



import pytest

from flink_rest_client import FlinkRestClient
from flink_rest_client.common import RestException
from flink_rest_client.v1.client import FlinkRestClientV1


class TestClient:

    def test_default_port(self):
        client = FlinkRestClient.get("test-host")
        assert client.port == 8081

    def test_unknown_version(self):
        with pytest.raises(RestException):
            FlinkRestClient.get("test-host", version="v2")

    def test_client_v1(self):
        client = FlinkRestClient.get("test-host", 8082)
        assert isinstance(client, FlinkRestClientV1)
        assert client.host == "test-host"
        assert client.port == 8082



import importlib_resources

from tests.v1.test_base import TestBase


class TestJarClient(TestBase):

    def test_all(self, simple_client, requests_mock):
        requests_mock.get(f'{simple_client.jars.prefix}', json={
            'address': 'http://jobmanager:8081',
            'files': []})
        response = simple_client.jars.all()

        assert isinstance(response, dict)
        assert response['address'] == 'http://jobmanager:8081'

    def test_upload(self, simple_client, requests_mock):
        with importlib_resources.path("tests.data.sample_jars", "StateMachineExample.jar") as fpath:
            requests_mock.post(f'{simple_client.jars.prefix}/upload', json={
                'filename': '/tmp/flink-web-dc68d2c5/flink-web-upload/c17af8f2-_StateMachineExample.jar',
                'status': 'success'})
            response = simple_client.jars.upload(fpath)

        assert isinstance(response, dict)
        assert response['status'] == 'success'
        assert response['filename'] == '/tmp/flink-web-dc68d2c5/flink-web-upload/c17af8f2-_StateMachineExample.jar'

    def test_run(self, simple_client, requests_mock):
        jar_id = 'test_jar_id'
        job_id = "bdtg564"
        requests_mock.post(f'{simple_client.jars.prefix}/{jar_id}/run', json={
            'jobid': job_id
        })
        response = simple_client.jars.run(jar_id)

        assert isinstance(response, str)
        assert response == job_id

    def test_get_plan(self, simple_client, requests_mock):
        jar_id = 'test_jar_id'
        requests_mock.post(f'{simple_client.jars.prefix}/{jar_id}/plan', json={
            'plan': {
                'jid': 'test_jar_id', 'name': 'State machine job',
                'nodes': []
            }
        })
        response = simple_client.jars.get_plan(jar_id)

        assert isinstance(response, dict)
        assert response['jid'] == jar_id

    def test_upload_and_run(self, simple_client, requests_mock):
        with importlib_resources.path("tests.data.sample_jars", "StateMachineExample.jar") as fpath:
            jar_id = "c17af8f2-_StateMachineExample.jar"
            job_id = "bdtg564"
            requests_mock.post(f'{simple_client.jars.prefix}/upload', json={
                'filename': f'/tmp/flink-web-dc68d2c5/flink-web-upload/{jar_id}',
                'status': 'success'})
            requests_mock.post(f'{simple_client.jars.prefix}/{jar_id}/run', json={
                'jobid': job_id
            })

            response = simple_client.jars.upload_and_run(fpath)
            assert isinstance(response, str)
            assert response == job_id

    def test_delete(self, simple_client, requests_mock):
        jar_id = 'test_jar_id'
        requests_mock.delete(f'{simple_client.jars.prefix}/{jar_id}', json={})
        response = simple_client.jars.delete(jar_id)

        assert isinstance(response, bool)
        assert response is True



from flink_rest_client.common import _execute_rest_request, RestException


class JobTrigger:
    def __init__(self, prefix, type_name, job_id, trigger_id):
        self._prefix = prefix
        self._type_name = type_name
        self.job_id = job_id
        self.trigger_id = trigger_id

    @property
    def status(self):
        return _execute_rest_request(
            url=f"{self._prefix}/{self.job_id}/{self._type_name}/{self.trigger_id}"
        )


class JobVertexSubtaskClient:
    def __init__(self, prefix):
        """
        Constructor.

        Parameters
        ----------
        prefix: str
            REST API url prefix. It must contain the host, port pair.
        """
        self._prefix = prefix

    @property
    def prefix_url(self):
        return f"{self._prefix}/subtasks"

    def subtask_ids(self):
        """
        Returns the subtask identifiers.

        Returns
        -------
        list
            Positive integer list of subtask ids.
        """
        return [elem["subtask"] for elem in self.accumulators()["subtasks"]]

    def accumulators(self):
        """
        Returns all user-defined accumulators for all subtasks of a task.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid/accumulators

        Returns
        -------
        dict
            User-defined accumulators
        """
        return _execute_rest_request(url=f"{self.prefix_url}/accumulators")

    def metric_names(self):
        """
        Returns the supported metric names.

        Returns
        -------
        list
            List of metric names.
        """
        return [
            elem["id"]
            for elem in _execute_rest_request(url=f"{self.prefix_url}/metrics")
        ]

    def metrics(self, metric_names=None, agg_modes=None, subtask_ids=None):
        """
        Provides access to aggregated subtask metrics.
        By default it returns with all existing metric names.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid/subtasks/metrics

        Parameters
        ----------
        metric_names: list
            (optional) List of selected specific metric names. Default: <all metrics>

        agg_modes: list
            (optional) List of aggregation modes which should be calculated. Available aggregations are: "min, max,
            sum, avg". Default: <all modes>

        subtask_ids: list
            List of positive integers to select specific subtasks. The list of valid subtask ids is available through
            the subtask_ids() method. Default: <all subtasks>.

        Returns
        -------
        dict
            Key-value pairs of metrics.
        """

        if metric_names is None:
            metric_names = self.metric_names()

        supported_agg_modes = ["min", "max", "sum", "avg"]
        if agg_modes is None:
            agg_modes = supported_agg_modes
        if len(set(agg_modes).difference(set(supported_agg_modes))) > 0:
            raise RestException(
                f"The provided aggregation modes list contains invalid value. Supported aggregation "
                f"modes: {','.join(supported_agg_modes)}; given list: {','.join(agg_modes)}"
            )

        if subtask_ids is None:
            subtask_ids = self.subtask_ids()

        params = {
            "get": ",".join(metric_names),
            "agg": ",".join(agg_modes),
            "subtasks": ",".join([str(elem) for elem in subtask_ids]),
        }
        query_result = _execute_rest_request(
            url=f"{self.prefix_url}/metrics", params=params
        )

        result = {}
        for elem in query_result:
            metric_name = elem.pop("id")
            result[metric_name] = elem

        return result

    def get(self, subtask_id):
        """
        Returns details of the current or latest execution attempt of a subtask.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid/subtasks/:subtaskindex

        Parameters
        ----------
        subtask_id: int
            Positive integer value that identifies a subtask.

        Returns
        -------
        dict

        """
        return _execute_rest_request(url=f"{self.prefix_url}/{subtask_id}")

    def get_attempt(self, subtask_id, attempt_id=None):
        """
        Returns details of an execution attempt of a subtask. Multiple execution attempts happen in case of
        failure/recovery.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid/subtasks/:subtaskindex/attempts/:attempt

        Parameters
        ----------
        subtask_id: int
            Positive integer value that identifies a subtask.

        attempt_id: int
            (Optional) Positive integer value that identifies an execution attempt.
            Default: current execution attempt's id

        Returns
        -------
        dict
            Details of the selected attempt.
        """
        if attempt_id is None:
            return self.get(subtask_id)
        return _execute_rest_request(
            url=f"{self.prefix_url}/{subtask_id}/attempts/{attempt_id}"
        )

    def get_attempt_accumulators(self, subtask_id, attempt_id=None):
        """
        Returns the accumulators of an execution attempt of a subtask. Multiple execution attempts happen in case of
        failure/recovery.

        Parameters
        ----------
        subtask_id: int
            Positive integer value that identifies a subtask.

        attempt_id: int
            (Optional) Positive integer value that identifies an execution attempt.
            Default: current execution attempt's id


        Returns
        -------
        dict
            The accumulators of the selected execution attempt of a subtask.
        """
        if attempt_id is None:
            attempt_id = self.get(subtask_id)["attempt"]
        return _execute_rest_request(
            url=f"{self.prefix_url}/{subtask_id}/attempts/{attempt_id}/accumulators"
        )


class JobVertexClient:
    def __init__(self, prefix, job_id, vertex_id):
        """
        Constructor.

        Parameters
        ----------
        prefix: str
            REST API url prefix. It must contain the host, port pair.
        """
        self._prefix = prefix
        self.job_id = job_id
        self.vertex_id = vertex_id

    @property
    def prefix_url(self):
        return f"{self._prefix}/{self.job_id}/vertices/{self.vertex_id}"

    @property
    def subtasks(self):
        return JobVertexSubtaskClient(self.prefix_url)

    def details(self):
        """
        Returns details for a task, with a summary for each of its subtasks.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid

        Returns
        -------
        dict
            details for a task.
        """
        return _execute_rest_request(url=self.prefix_url)

    def backpressure(self):
        """
        Returns back-pressure information for a job, and may initiate back-pressure sampling if necessary.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid/backpressure

        Notes
        -----
        The deprecated status means that the back pressure stats are not available.

        Returns
        -------
        dict
            Backpressure information
        """
        return _execute_rest_request(url=f"{self.prefix_url}/backpressure")

    def metric_names(self):
        """
        Returns the supported metric names.

        Returns
        -------
        list
            List of metric names.
        """
        return [
            elem["id"]
            for elem in _execute_rest_request(url=f"{self.prefix_url}/metrics")
        ]

    def metrics(self, metric_names=None):
        """
        Provides access to task metrics.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid/metrics

        Returns
        -------
        dict
            Task metrics.
        """
        if metric_names is None:
            metric_names = self.metric_names()

        params = {"get": ",".join(metric_names)}
        query_result = _execute_rest_request(
            url=f"{self.prefix_url}/metrics", params=params
        )
        result = {}
        for elem in query_result:
            metric_name = elem.pop("id")
            result[metric_name] = elem["value"]
        return result

    def subtasktimes(self):
        """
        Returns time-related information for all subtasks of a task.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid/subtasktimes

        Returns
        -------
        dict
            Time-related information for all subtasks
        """
        return _execute_rest_request(url=f"{self.prefix_url}/subtasktimes")

    def taskmanagers(self):
        """
        Returns task information aggregated by task manager.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid/taskmanagers

        Returns
        -------
        dict
            Task information aggregated by task manager.
        """
        return _execute_rest_request(url=f"{self.prefix_url}/taskmanagers")

    def watermarks(self):
        """
        Returns the watermarks for all subtasks of a task.

        Endpoint: [GET] /jobs/:jobid/vertices/:vertexid/watermarks

        Returns
        -------
        list
            Watermarks for all subtasks of a task.
        """
        return _execute_rest_request(url=f"{self.prefix_url}/watermarks")


class JobsClient:
    def __init__(self, prefix):
        """
        Constructor.

        Parameters
        ----------
        prefix: str
            REST API url prefix. It must contain the host, port pair.
        """
        self.prefix = f"{prefix}/jobs"

    def all(self):
        """
        Returns an overview over all jobs and their current state.

        Endpoint: [GET] /jobs

        Returns
        -------
        list
            List of jobs and their current state.
        """
        return _execute_rest_request(url=self.prefix)["jobs"]

    def job_ids(self):
        """
        Returns the list of job_ids.

        Returns
        -------
        list
            List of job ids.
        """
        return [elem["id"] for elem in self.all()]

    def overview(self):
        """
        Returns an overview over all jobs.

        Endpoint: [GET] /jobs/overview

        Returns
        -------
        list
            List of existing jobs.
        """
        return _execute_rest_request(url=f"{self.prefix}/overview")["jobs"]

    def metric_names(self):
        """
        Returns the supported metric names.

        Returns
        -------
        list
            List of metric names.
        """
        return [
            elem["id"] for elem in _execute_rest_request(url=f"{self.prefix}/metrics")
        ]

    def metrics(self, metric_names=None, agg_modes=None, job_ids=None):
        """
        Returns an overview over all jobs.

        Endpoint: [GET] /jobs/metrics

        Parameters
        ----------
        metric_names: list
            (optional) List of selected specific metric names. Default: <all metrics>

        agg_modes: list
            (optional) List of aggregation modes which should be calculated. Available aggregations are: "min, max,
            sum, avg". Default: <all modes>

        job_ids: list
            List of 32-character hexadecimal strings to select specific jobs. The list of valid jobs
            are available through the job_ids() method. Default: <all taskmanagers>.


        Returns
        -------
        dict
            Aggregated job metrics.
        """
        if metric_names is None:
            metric_names = self.metric_names()

        supported_agg_modes = ["min", "max", "sum", "avg"]
        if agg_modes is None:
            agg_modes = supported_agg_modes
        if len(set(agg_modes).difference(set(supported_agg_modes))) > 0:
            raise RestException(
                f"The provided aggregation modes list contains invalid value. Supported aggregation "
                f"modes: {','.join(supported_agg_modes)}; given list: {','.join(agg_modes)}"
            )

        if job_ids is None:
            job_ids = self.job_ids()

        params = {
            "get": ",".join(metric_names),
            "agg": ",".join(agg_modes),
            "jobs": ",".join(job_ids),
        }
        query_result = _execute_rest_request(
            url=f"{self.prefix}/metrics", params=params
        )

        result = {}
        for elem in query_result:
            metric_name = elem.pop("id")
            result[metric_name] = elem

        return result

    def get(self, job_id):
        """
        Returns details of a job.

        Endpoint: [GET] /jobs/:jobid

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        dict
            Details of the selected job.
        """
        return _execute_rest_request(url=f"{self.prefix}/{job_id}")

    def get_config(self, job_id):
        """
        Returns the configuration of a job.

        Endpoint: [GET] /jobs/:jobid/config

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        dict
            Job configuration
        """
        return _execute_rest_request(url=f"{self.prefix}/{job_id}/config")

    def get_exceptions(self, job_id):
        """
        Returns the most recent exceptions that have been handled by Flink for this job.

        Endpoint: [GET] /jobs/:jobid/exceptions

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        dict
            The most recent exceptions.
        """
        return _execute_rest_request(url=f"{self.prefix}/{job_id}/exceptions")

    def get_execution_result(self, job_id):
        """
        Returns the result of a job execution. Gives access to the execution time of the job and to all accumulators
        created by this job.

        Endpoint: [GET] /jobs/:jobid/execution-result

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        dict
            The execution result of the selected job.
        """
        return _execute_rest_request(url=f"{self.prefix}/{job_id}/execution-result")

    def get_metrics(self, job_id, metric_names=None):
        """
        Provides access to job metrics.

        Endpoint: [GET] /jobs/:jobid/metrics

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        metric_names: list
            (optional) List of selected specific metric names. Default: <all metrics>

        Returns
        -------
        dict
            Job metrics.
        """
        if metric_names is None:
            metric_names = self.metric_names()
        params = {"get": ",".join(metric_names)}
        query_result = _execute_rest_request(
            url=f"{self.prefix}/{job_id}/metrics", params=params
        )
        return dict([(elem["id"], elem["value"]) for elem in query_result])

    def get_plan(self, job_id):
        """
        Returns the dataflow plan of a job.

        Endpoint: [GET] /jobs/:jobid/plan

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        dict
            Dataflow plan
        """
        return _execute_rest_request(url=f"{self.prefix}/{job_id}/plan")["plan"]

    def get_vertex_ids(self, job_id):
        """
        Returns the ids of vertices of the selected job.

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        list
            List of identifiers.
        """
        return [elem["id"] for elem in self.get(job_id)["vertices"]]

    def get_accumulators(self, job_id, include_serialized_value=None):
        """
        Returns the accumulators for all tasks of a job, aggregated across the respective subtasks.

        Endpoint: [GET] /jobs/:jobid/accumulators

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        include_serialized_value: bool
             (Optional) Boolean value that specifies whether serialized user task accumulators should be included in
             the response.

        Returns
        -------
        dict
            Accumulators for all task.
        """

        params = {}
        if include_serialized_value is not None:
            params["includeSerializedValue"] = (
                "true" if include_serialized_value else "false"
            )

        return _execute_rest_request(
            url=f"{self.prefix}/{job_id}/accumulators", http_method="GET", params=params
        )

    def get_checkpointing_configuration(self, job_id):
        """
        Returns the checkpointing configuration of the selected job_id

        Endpoint: [GET] /jobs/:jobid/checkpoints/config

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        dict
            Checkpointing configuration of the selected job.
        """
        return _execute_rest_request(
            url=f"{self.prefix}/{job_id}/checkpoints/config", http_method="GET"
        )

    def get_checkpoints(self, job_id):
        """
        Returns checkpointing statistics for a job.

        Endpoint: [GET] /jobs/:jobid/checkpoints

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        dict
            Checkpointing statistics for the selected job: counts, summary, latest and history.
        """
        return _execute_rest_request(
            url=f"{self.prefix}/{job_id}/checkpoints", http_method="GET"
        )

    def get_checkpoint_ids(self, job_id):
        """
        Returns checkpoint ids of the job_id.

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        list
            List of checkpoint ids.
        """
        return [elem["id"] for elem in self.get_checkpoints(job_id=job_id)["history"]]

    def get_checkpoint_details(self, job_id, checkpoint_id, show_subtasks=False):
        """
        Returns details for a checkpoint.

        Endpoint: [GET] /jobs/:jobid/checkpoints/details/:checkpointid

        If show_subtasks is true:
        Endpoint: [GET] /jobs/:jobid/checkpoints/details/:checkpointid/subtasks/:vertexid

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        checkpoint_id: int
            Long value that identifies a checkpoint.

        show_subtasks: bool
            If it is True, the details of the subtask are also returned.

        Returns
        -------
        dict

        """
        checkpoint_details = _execute_rest_request(
            url=f"{self.prefix}/{job_id}/checkpoints/details/{checkpoint_id}",
            http_method="GET",
        )
        if not show_subtasks:
            return checkpoint_details

        subtasks = {}
        for vertex_id in checkpoint_details["tasks"].keys():
            subtasks[vertex_id] = _execute_rest_request(
                url=f"{self.prefix}/{job_id}/checkpoints/details/{checkpoint_id}/subtasks/{vertex_id}",
                http_method="GET",
            )
        checkpoint_details["subtasks"] = subtasks
        return checkpoint_details

    def rescale(self, job_id, parallelism):
        """
        Triggers the rescaling of a job. This async operation would return a 'triggerid' for further query identifier.

        Endpoint: [GET] /jobs/:jobid/rescaling

        Notes
        -----
        Using Flink version 1.12, the method will raise RestHandlerException because this rescaling is temporarily
        disabled. See FLINK-12312.

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.
        parallelism: int
            Positive integer value that specifies the desired parallelism.

        Returns
        -------
        JobTrigger
            Object that can be used to query the status of rescaling.
        """
        params = {"parallelism": parallelism}
        trigger_id = _execute_rest_request(
            url=f"{self.prefix}/{job_id}/rescaling", http_method="PATCH", params=params
        )["triggerid"]
        return JobTrigger(self.prefix, "rescaling", job_id, trigger_id)

    def create_savepoint(self, job_id, target_directory, cancel_job=False):
        """
        Triggers a savepoint, and optionally cancels the job afterwards. This async operation would return a
        JobTrigger for further query identifier.

        Endpoint: [GET] /jobs/:jobid/savepoints

        Notes
        -----
        The target directory has to be a location accessible by both the JobManager(s) and TaskManager(s)
        e.g. a location on a distributed file-system or Object Store.

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.
        target_directory: str
            Savepoint target directory.
        cancel_job: bool
            If it is True, it also stops the job after the savepoint creation.

        Returns
        -------
        JobTrigger
            Object that can be used to query the status of savepoint.
        """
        trigger_id = _execute_rest_request(
            url=f"{self.prefix}/{job_id}/savepoints",
            http_method="POST",
            accepted_status_code=202,
            json={"cancel-job": cancel_job, "target-directory": target_directory},
        )["request-id"]
        return JobTrigger(self.prefix, "savepoints", job_id, trigger_id)

    def terminate(self, job_id):
        """
        Terminates a job.

        Endpoint: [PATCH] /jobs/:jobid

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.

        Returns
        -------
        bool
            True if the job has been canceled, otherwise False.
        """
        res = _execute_rest_request(
            url=f"{self.prefix}/{job_id}", http_method="PATCH", accepted_status_code=202
        )
        if len(res) < 1:
            return True
        else:
            return False

    def stop(self, job_id, target_directory, drain=False):
        """
        Stops a job with a savepoint. This async operation would return a JobTrigger for further query identifier.

        Attention: The target directory has to be a location accessible by both the JobManager(s) and TaskManager(s)
        e.g. a location on a distributed file-system or Object Store.

        Draining emits the maximum watermark before stopping the job. When the watermark is emitted, all event time
        timers will fire, allowing you to process events that depend on this timer (e.g. time windows or process
        functions). This is useful when you want to fully shut down your job without leaving any unhandled events
        or state.

        Endpoint: [GET] /jobs/:jobid/stop

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.
        target_directory: str
            Savepoint target directory.
        drain: bool
            (Optional) If it is True, it emits the maximum watermark before stopping the job. default: False

        Returns
        -------
        JobTrigger
            Object that can be used to query the status of savepoint.
        """
        data = {
            "drain": False if drain is None else drain,
            "targetDirectory": target_directory,
        }

        trigger_id = _execute_rest_request(
            url=f"{self.prefix}/{job_id}/stop",
            http_method="POST",
            accepted_status_code=202,
            json=data,
        )["request-id"]
        return JobTrigger(self.prefix, "savepoints", job_id, trigger_id)

    def get_vertex(self, job_id, vertex_id):
        """
        Returns a JobVertexClient.

        Parameters
        ----------
        job_id: str
            32-character hexadecimal string value that identifies a job.
        vertex_id: str
            32-character hexadecimal string value that identifies a vertex.

        Returns
        -------
        JobVertexClient
            JobVertexClient instance that can execute vertex related queries.
        """
        return JobVertexClient(self.prefix, job_id, vertex_id)



import pytest

from flink_rest_client.v1.client import FlinkRestClientV1


class TestBase:

    @pytest.fixture(scope="class")
    def simple_client(self):
        return FlinkRestClientV1("host", 8081)



import math

from flink_rest_client.v1.jobs import JobVertexSubtaskClient
from tests.v1.test_base import TestBase


class TestJobsClient(TestBase):

    def test_subtasks(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        response = simple_client.jobs.get_vertex(jid, vid).subtasks

        assert isinstance(response, JobVertexSubtaskClient)

    def test_details(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        vc = simple_client.jobs.get_vertex(jid, vid)

        requests_mock.get(f'{vc.prefix_url}', json={
            'id': 'bc764cd8ddf7a0cff126f51c16239658',
            'name': 'Source: Custom Source',
            'parallelism': 1
        })
        response = vc.details()

        assert isinstance(response, dict)
        assert response['id'] == 'bc764cd8ddf7a0cff126f51c16239658'
        assert response['name'] == 'Source: Custom Source'
        assert response['parallelism'] == 1

    def test_backpressure(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        vc = simple_client.jobs.get_vertex(jid, vid)

        requests_mock.get(f'{vc.prefix_url}/backpressure', json={'status': 'deprecated'})
        response = vc.backpressure()

        assert isinstance(response, dict)
        assert response['status'] == 'deprecated'

    def test_metric_names(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        vc = simple_client.jobs.get_vertex(jid, vid)

        requests_mock.get(f'{vc.prefix_url}/metrics', json=[
            {'id': '0.numBytesInRemote'},
            {'id': '0.Source__Custom_Source.numRecordsIn'},
            {'id': '0.numBytesInRemotePerSecond'},
        ])
        response = vc.metric_names()

        assert isinstance(response, list)
        assert len(response) == 3
        assert response[0] == '0.numBytesInRemote'
        assert response[1] == '0.Source__Custom_Source.numRecordsIn'
        assert response[2] == '0.numBytesInRemotePerSecond'

    def test_metrics(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        vc = simple_client.jobs.get_vertex(jid, vid)

        requests_mock.get(f'{vc.prefix_url}/metrics', json=[
            {'id': '0.numBytesInRemote', 'value': 0.0},
            {'id': '0.Source__Custom_Source.numRecordsIn', 'value': 40.0},
        ])
        response = vc.metrics(['0.numBytesInRemote', '0.Source__Custom_Source.numRecordsIn'])

        assert isinstance(response, dict)
        assert math.isclose(response['0.numBytesInRemote'], 0.0)
        assert math.isclose(response['0.Source__Custom_Source.numRecordsIn'], 40.0)

    def test_subtasktimes(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        vc = simple_client.jobs.get_vertex(jid, vid)

        requests_mock.get(f'{vc.prefix_url}/subtasktimes', json={
            'id': 'bc764cd8ddf7a0cff126f51c16239658',
            'name': 'Source: Custom Source',
            'now': 101010,
            'subtasks': [{'subtask': 0,
                          'host': 'flink-taskmanager', 'duration': 5127315,
                          'timestamps': {'SCHEDULED': 1625895875, 'CANCELED': 0, 'FINISHED': 0, 'RECONCILING': 0,
                                         'CREATED': 124815827, 'RUNNING': 1215896082, 'CANCELING': 0,
                                         'FAILED': 0, 'DEPLOYING': 6281535}}]})
        response = vc.subtasktimes()

        assert isinstance(response, dict)
        assert response['subtasks'][0]['timestamps']['SCHEDULED'] == 1625895875

    def test_watermarks(self, simple_client, requests_mock):
        jid = "efjbcepigfvrui235324ui"
        vid = "12hfkbro943rri235324ui"
        vc = simple_client.jobs.get_vertex(jid, vid)

        requests_mock.get(f'{vc.prefix_url}/watermarks', json=[])
        response = vc.watermarks()

        assert isinstance(response, list)
        assert len(response) == 0

